/*
 * array.c - array interface functions
 * 
 * include LICENSE
 */
#include <stdio.h>
#include <string.h>

#include <strmem.h>
#include <array.h>

#ifdef TRACE_MEM
#include <tracemem.h>
#endif

/*
 *** \brief Allocates memory for a new Array object.
 */


Array *array_strPtr_new( int count )
{
   Array *ary = array_new( 0, count );

#ifdef TRACE_MEM
   ary->destroyFunc = str_free_func;
#else
    ary->destroyFunc = app_free;
#endif
   return ary;
}

Array *array_new( int elSize, int count )
{
   Array *ary;

   ary =  app_new0(Array, 1);
   array_construct( ary, elSize, count );
   app_class_overload_destroy( (AppClass *) ary, array_destroy );
   return ary;
}

/** \brief Constructor for the Array object. */

void array_construct( Array *ary, int elSize, int count )
{
   app_class_construct( (AppClass *) ary );

   if (count == 0 ){
      count = DEF_ELEM_COUNT;
   } else if ( count < 0 ){
      count = app_power_of_2(count);
   }
   if (elSize == 0 ){
      elSize = sizeof(char *); /* sizeof pointer */
   }

   ary->nAlloc = count;
   ary->elSize = elSize;
   ary->ptr = app_malloc0 ( elSize * count );
   /*
    *  default function to destroy an element
    *  Must be overloaded for non AppClass element : char *, dlist *
    */
   ary->destroyFunc = (ElmDestroyFP_func) app_class_unref ;
}

/** \brief Destructor for the Array object. */

void array_destroy(void *ary)
{
   Array *this = (Array *) ary;

   if (ary == NULL) {
      return;
   }

   array_tbl_delete_all(this);
   
   app_free(this->ptr);
   app_class_destroy( ary );
}

void array_tbl_delete_all(Array *ary)
{
   int i;
   void **pptr;
   ElmDestroyFP_func func = ary->destroyFunc;
   
   if ( func ) {
      for ( i = 0 ; i < ary->nelem ; i++ ){
         pptr = ary->ptr + (i * ary->elSize);
         if ( *pptr ){
            func ( *pptr );
         }
      }
   }
   ary->nelem = 0;
}

void array_set_elem_destroyFunc(Array *ary, ElmDestroyFP_func func)
{
   ary->destroyFunc = (ElmDestroyFP_func) func;
}

void array_tbl_delete(Array *ary, int index)
{
   if ( index >= ary->nelem) {
      msg_error ( "Index %d > Nb elem %d", index, ary->nelem);
      return;
   }
   void **pptr = (void **) ary->ptr;
   if ( pptr[index] ){
      ary->destroyFunc ( pptr[index] );
   }
   memcpy ( &pptr[index], &pptr[index + 1], ary->elSize * (ary->nelem - index));
   ary->nelem--;
}

void *array_table_get(Array *ary)
{
   return ary->ptr ;
}

void *array_tbl_get(Array *ary, int index)
{
   if ( index >= ary->nelem) {
      msg_error ( "Index %d > Nb elem %d", index, ary->nelem);
      return NULL;
   }
   void **pptr = (void **) ary->ptr;
   return pptr[index] ;
}

/*
 * get the element and remove it from array without destroying the element
 */
void *array_tbl_remove(Array *ary, int index)
{
   if ( index >= ary->nelem) {
      msg_error ( "Index %d > Nb elem %d", index, ary->nelem);
      return NULL;
   }
   void **pptr = (void **) ary->ptr;
   void *ptr = pptr[index] ;
   memcpy ( &pptr[index], &pptr[index + 1], ary->elSize * (ary->nelem - index));
   ary->nelem--;
   return ptr ;
}
   
void array_tbl_expand(Array *ary, int n)
{
   if ( n > ary->nAlloc ) {
      ary->nAlloc <<= 1;
      ary->ptr = app_realloc(ary->ptr, ary->elSize * ary->nAlloc);
      memset(ary->ptr + (ary->nelem * ary->elSize), 0,
             ary->elSize * ( ary->nAlloc - ary->nelem) );
   }
}

void array_tbl_add(Array *ary, void *elPtr)
{
   array_tbl_expand(ary, ary->nelem + 1);
   void **pptr = (void **) ary->ptr;
   pptr[ary->nelem++] = elPtr;
}

/*
 * strdup a string and add it
 */
void array_strPtr_add(Array *ary, char *str )
{
   char *newstr = app_strdup(str);
   array_tbl_add(ary, newstr);
}

/*
 *  insert before index
 */
void array_tbl_insert(Array *ary, void *elPtr, int index)
{
   if ( index > ary->nelem) {
      msg_error ( "Index %d > Nb elem %d", index, ary->nelem);
      return;
   }
      
   array_tbl_expand(ary, ary->nelem + 1);
   void **pptr = (void **) ary->ptr;
   memmove ( &pptr[index + 1], &pptr[index], ary->elSize * (ary->nelem - index));
   pptr[index] = elPtr;
   ary->nelem++;
}

/*
 * strdup a string and insert it
 */
void array_strPtr_insert(Array *ary, char *str, int index )
{
   char *newstr = app_strdup(str);
   array_tbl_insert(ary, newstr, index);
}

/*
 *  replace element at index
 *  return the old element
 */
void *array_tbl_replace(Array *ary, void *elPtr, int index)
{
   if ( index > ary->nelem) {
      msg_error ( "Index %d > Nb elem %d", index, ary->nelem);
      return NULL;
   }
      
   void **pptr = (void **) ary->ptr;
   void *ptr = pptr[index] ;
   pptr[index] = elPtr;
   return ptr ;
}

/*
 * replace the element with the new one
 *  and kill the old one
 */
void array_tbl_replace_kill(Array *ary, void *elPtr, int index)
{
   void *ptr = array_tbl_replace(ary, elPtr, index) ;
   if ( ptr ) {
      ary->destroyFunc ( ptr ) ;
   }
}

/*
 * strdup a string and replace it at index
 */
char *array_strPtr_replace(Array *ary, char *str, int index )
{
   char *newstr = app_strdup(str);
   return (char *) array_tbl_replace(ary, newstr, index);
}

void array_strPtr_replace_kill(Array *ary, char *str, int index )
{
   char *oldstr = array_strPtr_replace(ary, str, index );
   if ( oldstr ) {
      ary->destroyFunc ( oldstr ) ;
   }
}

int array_tbl_index(Array *ary, void *elPtr)
{
   int i;
   void **pptr = (void **) ary->ptr;
   
   for ( i = 0 ; i < ary->nelem ; i++ ){
      if (  pptr[i] == elPtr ) {
	 return i;
      }
   }
   return -1;
}

void array_set_nelem (Array *ary, int nelem)
{
   ary->nelem = nelem;
   return ;
}

int array_get_nelem (Array *ary)
{
   return ary->nelem ;
}

int array_strPtr_cmp( AppClass *d1, AppClass *d2 )
{
   return (app_strcmp( ((char *) d1), ((char *) d2) ) );
}

#if 0
/*
 * return 1 if the string is found in array
 */
int array_strPtr_lookup (Array *ary, char *str)
{
   int i;
   
   for ( i = 0 ; i < array_get_nelem (ary) ; i++ ) {
      if ( app_strcmp(array_strPtr_get(ary, i), str) == 0 ) { 
	 return 1;
      }
   }
   return 0;
}
#endif

/** \brief Lookup an item in the array.
 *
 *  Walk the array pointed to by ary and return a pointer to the data if
 *  found. If not found, return NULL. 
 *
 *  \param ary The array to be iterated.
 *  \param data The data to be passed to the func when it is applied.
 *  \param cmp  A function to be used for comparing the items.
 *
 *  \return     A pointer to the data found, or NULL if not found.
 *
 *  Warning: element list Must be sorted before creating the table
 */
AppClass *array_lookup_sorted( Array *ary, AppClass *data, ArrayFP_Cmp cmp, int *index  )
{
   int i;
   int ret;
   int lo ;
   int hi ;
   AppClass *entry ;
   
   if (cmp == NULL) {
      msg_fatal( "compare function not specified" );
   }
   lo = 0;
   hi =  ary->nelem;

   while (lo != hi) {
      i = (hi + lo)  / 2 ;

      entry = (AppClass *) array_tbl_get( ary, i );
      ret = (*cmp)(entry, data);
      if ( ret > 0 ) {  /*  entry->symbol > symbol */
         hi = i;
      } else if ( ret < 0 ) {  /*  entry->symbol < symbol */
         lo = i + 1;
      } else {         /*  entry->symbol == symbol  */
         /* symbol found */
         if ( index ){
            *index = i;
         }
         return entry;
      }
   }
   return NULL;
}

/*
 * lookup an unsorted list
 */
AppClass *array_lookup( Array *ary, AppClass *data, ArrayFP_Cmp cmp, int *index )
{
   int i;
   AppClass *entry ;
   
   if (cmp == NULL) {
      msg_fatal( "compare function not specified" );
   }

   for ( i = 0; i < ary->nelem ; i++) {
      entry = (AppClass *) array_tbl_get( ary, i );
      if ( entry && (*cmp)(entry, data) == 0 ) {  /*  entry->symbol == symbol */
         /* symbol found */
         if ( index ){
            *index = i;
         }
         return entry;
      }
   }
   return NULL;
}

void array_sort(Array *ary, __compar_fn_t cmp_func )
{
   qsort( ary->ptr, ary->nelem, ary->elSize, cmp_func);
}
